import re
from typing import Any, Optional

def __get_entity(pattern: str, string: str) -> Optional[str]:
    r = re.search(pattern, string)
    return r.group() if r is not None else None

def extract_date_time(content: str, instructions: list, prompt_engine: Any) -> dict:
    """Extract time information from return value of function call extract_date_time

    Args:
        content (str): return value from extract_date_time fucntion call

    Returns:
        dict: start time, end time, and current time
    """
    start_time = r"(?<=start time:\s)\d{4}-\d{2}-\d{2}\s\d{2}:\d{2}:\d{2}.*(?=Ext)"
    end_time = r"(?<=end time:\s)\d{4}-\d{2}-\d{2}\s\d{2}:\d{2}:\d{2}.*(?=Ext)"
    current_time = r"(?<=Current time:\s)\d{4}-\d{2}-\d{2}\s\d{2}:\d{2}:\d{2}"

    return {
        'name': 'extract_date_time',
        'parameters': {
            'start_time': __get_entity(start_time, content),
            'end_time': __get_entity(end_time, content),
            'current_time': __get_entity(current_time, content),
        },
        'role': 'function',
    }

def extract_coordinates(content: str, instructions: list, prompt_engine: Any) -> dict:
    """
    Extract latitude and longitude information from return value of function call extract_coordinatess

    Args:
        content (str): return value from extract_coordinates fucntion call

    Returns:
        dict: latitude and longitude
    """

    latitude = r"(?<=latitude:\s)-?\d{1,3}\.\d{1,6}"
    longitude = r"(?<=longitude:\s)-?\d{1,3}\.\d{1,6}"

    return {
        'name': 'extract_coordinates',
        'parameters': {
            'latitude': __get_entity(latitude, content),
            'longitude': __get_entity(longitude, content),
        },
        'role': 'function',
    }

def get_movie_theaters(content: str, instructions: list, prompt_engine: Any) -> dict:
    """
    Get list of movie theaters given coordinates

    Args:
        content (str): return value from get_movie_theaters fucntion call

    Returns:
        dict:
            reference_number,
            popular_theaters,
            recent_visited_theaters,
            user_preferred_theaters,
            requested_location_based_theaters
    """

    reference_number = r"(?<=reference number:\s)ref_\d+"
    popular_theaters = r"(?<=Popular theaters:\s).*"
    recent_visited_theaters = r"(?<=Recently visited theaters:\s).*"
    user_preferred_theaters = r"(?<=User-preferred theaters:\s).*"
    requested_location_based_theaters = r"(?<=Requested location-based theaters:\s).*"

    return {
        'name': 'get_movie_theaters',
        'parameters': {
            'reference_number': __get_entity(reference_number, content),
            'popular_theaters': __get_entity(popular_theaters, content),
            'popular_theaters': __get_entity(popular_theaters, content),
            'user_preferred_theaters': __get_entity(user_preferred_theaters, content),
            'requested_location_based_theaters': __get_entity(requested_location_based_theaters, content),
        },
        'role': 'function',
    }

def get_movie_titles(content: str, instructions: list, prompt_engine: Any) -> dict:
    """
    Get list of movie titles given coordinates

    Args:
        content (str): return value from get_movie_titles fucntion call

    Returns:
        dict:
            reference_number,
            movie_titles,
    """

    reference_number = r"(?<=reference number:\s)ref_\d+"
    movie_titles = r'(?:(?<=reference number:\sref_\d{3}\)\s)|(?<=reference number:\sref_\d{4}\)\s)).*'

    return {
        'name': 'get_movie_titles',
        'parameters': {
            'reference_number': __get_entity(reference_number, content),
            'movie_titles': __get_entity(movie_titles, content),
        },
        'role': 'function',
    }

def get_movie_schedules(content: str, instructions: list, prompt_engine: Any) -> dict:
    """
    Get list of movie schedules given coordinates

    Args:
        content (str): return value from get_movie_schedules fucntion call

    Returns:
        dict:
            reference_number,
            movie_schedules,
    """

    reference_number = r"(?<=reference number:\s)ref_\d+"
    movie_title = r'(?<=-\sScreening\sschedule\sof\s").*(?=" movie\sin)'
    movie_theater = r'(?<=movie\sin\s").*(?=" theater:)'
    movie_schedules = r'(?<=" theater:).*'

    return {
        'name': 'get_movie_schedules',
        'parameters': {
            'reference_number': __get_entity(reference_number, content),
            'movie_title': __get_entity(movie_title, content),
            'movie_theater': __get_entity(movie_theater, content),
            'movie_schedules': __get_entity(movie_schedules, content),
        },
        'role': 'function',
    }

def book_movie_tickets(content: str, instructions: list, prompt_engine: Any) -> dict:
    return {
        'name': 'book_movie_tickets',
        'parameters': {},
        'role': 'function',
    }

def search_movie_information(content: str, instructions: list, prompt_engine: Any) -> dict:
    movie_title = r"(?<='title':\s).*(?=\s)"
    movie_snippet = r"(?<=snippet:\s).*"

    content_list = content.split('\n')
    movies = []
    for i in range(1, len(content_list), 2):
        title = __get_entity(movie_title, content_list[i])
        snippet = __get_entity(movie_snippet, content_list[i + 1])
        movies.append({'title': title, 'snippet': snippet})

    return {
        'name': 'search_movie_information',
        'parameters': movies,
        'role': 'function',
    }

def search_benefit_for_booking_movie(content: str, instructions: list, prompt_engine: Any) -> dict:
    return {
        'name': 'search_benefit_for_booking_movie',
        'parameters': {
            'benefit': content,
        },
        'role': 'function',
    }



#
# Function definition
#

EXTRACT_DATE_TIME = {
    'required': ['query'],
    'parameters': {
        'query': {
            'name': 'query',
            'equivalence': {
                'tester': 'get_equivalence_tester',
                'category': 'date, time, time range, date range',
            },
            'consistency': {
                'tester': 'get_category_consistency_tester',
                'instruction': 'latest',
                'category': 'date, time, time range, date range',
            },
            'grammar': {
                'tester': 'get_grammar_tester',
            },
        }
    }
}

EXTRACT_COORDINATES = {
    'required': ['query'],
    'parameters': {
        'query': {
            'name': 'query',
            'equivalence': {
                'tester': 'get_equivalence_tester',
                'category': 'date, time, time range, date range',
            },
            'consistency': {
                'tester': 'get_category_consistency_tester',
                'instruction': 'latest',
                'category': 'place, land mark, district, building',
            },
            'grammar': {
                'tester': 'get_grammar_tester',
            },
        }
    }
}

GET_MOVIE_THEATERS = {
    'required': ['latitude', 'longitude'],
    'parameters': {
        'latitude': {
            'name': 'latitude',
            'equivalence': {
                'tester': 'get_equivalence_tester',
                'category': 'coordinate',
            },
            'consistency': {
                'tester': 'get_number_consistency_tester',
                'instruction': 'latest',
                'range': [-90, 90],
            },
        },
        'longitude': {
            'name': 'longitude',
            'equivalence': {
                'tester': 'get_equivalence_tester',
                'category': 'coordinate',
            },
            'consistency': {
                'tester': 'get_number_consistency_tester',
                'instruction': 'latest',
                'range': [-180, 180],
            },
        },
    },
}

GET_MOVIE_TITLES = {
    'required' : [],
    'parameters': {
        'start_time': {
            'name': 'start_time',
            'equivalence': {
                'tester': 'get_equivalence_tester',
                'category': 'datetime, YYYY-mm-dd HH:MM:SS',
            },
            'consistency': {
                'tester': 'get_category_consistency_tester',
                'category': 'datetime, YYYY-mm-dd HH:MM:SS',
                'comment': 'If the start time is not provided explicitly, it should be inferred from the compound of the current time and user request.',
                'instruction': 'all',
            },
        },
        'end_time': {
            'name': 'end_time',
            'equivalence': {
                'tester': 'get_equivalence_tester',
                'category': 'datetime, YYYY-mm-dd HH:MM:SS',
            },
            'consistency': {
                'tester': 'get_category_consistency_tester',
                'category': 'datetime, YYYY-mm-dd HH:MM:SS',
                'comment': 'If the end time is not provided explicitly, it should be inferred from the compound of the current time and user request.',
                'instruction': 'all',
            },
        },
        'movie_theater': {
            'name': 'movie_theater',
            'equivalence': {
                'tester': 'get_equivalence_tester',
                'category': 'theater name',
            },
            'consistency': {
                'tester': 'get_category_consistency_tester',
                'category': 'theater name',
                'instruction': 'all',
            },
        },
        'query': {
            'name': 'query',
            'equivalence': {
                'tester': 'get_equivalence_tester',
                'category': 'movie title',
            },
            'consistency': {
                'tester': 'get_category_consistency_tester',
                'category': 'movie title',
                'instruction': 'latest',
            },
            'grammar': {
                'tester': 'get_grammar_tester',
            },
        },
    },
}



import json
import itertools
import logging
import re
from typing import Optional, Union

from tqdm import tqdm

from langchain_openai import ChatOpenAI

from equivalent_llm import function, function_call, reservation, user

class EquvalentLLM:

    configurations = {
        'temperature': 0.7,
        'model': 'gpt-4-0125-preview',
        'prompt_engine': 'OpenAI',
        'parse_instruction': True,
        'test_consistency': True,
        'test_grammar': True,
        'test_elegance': True,
        'api': 'v1',
    }

    # initialization
    instructions = []
    instruction: dict = {}
    parsed_instructions: list = []
    parsed_reference = None
    parsed_generated = None

    def __load_open_ai(self) -> ChatOpenAI:
        return ChatOpenAI(
            temperature = self.configurations['temperature'],
            model = self.configurations['model']
        )

    def __json_loads(self, json_str: str) -> dict:
        parsed = json.loads(json_str)
        #if isinstance(parsed, list) and len(parsed) == 1:
        if isinstance(parsed, list):
            return parsed[0]
        else:
            return parsed

    def __parse_preknowledge(self, preknowledge: str) -> None:
        contents = preknowledge.replace('\n','').split('</s>')
        messages = [
            list(filter(lambda x: x is not None and x.strip() != '',
                re.sub('.*<</SYS>>', '', content)
                .replace('<s>[INST]', '')
                .split('[/INST]')))
            for content in contents
        ]
        self.instructions = list(itertools.chain([ self.__json_loads(m) for message in messages[:-1] for m in message ]))
        self.instruction = self.__json_loads(messages[-1][0])
        self.instructions.append(self.instruction)

    def __init__(self, preknowledge, reference, generated, logger = None, **kwargs):
        self.preknowledge = preknowledge
        self.reference = reference
        self.generated = generated

        self.logger = logger if logger is not None else logging.getLogger(__name__)

        self.configurations.update(kwargs)
        if self.configurations['prompt_engine'] == 'OpenAI':
            self.prompt_engine = self.__load_open_ai()
        else:
            raise ValueError(f"Unknown prompt engine: {self.configurations['prompt_engine']}")

        self.__parse_preknowledge(self.preknowledge)

    def update_configurations(self, **kwargs) -> None:
        self.configurations.update(kwargs)
        self.logger.debug(self.configurations)

    def parse_instructions(self) -> None:
        parsed_instructions = []
        for index, instruction in enumerate(self.instructions):
            match instruction.get('role'):
                case 'user':
                    parsed_instructions.append(
                        user.get_entities(instruction['content'], parsed_instructions[0:index], self.prompt_engine, self.logger)
                    )
                case 'function':
                    parser = getattr(function, instruction['name'])
                    parsed_instructions.append(parser(instruction['content'], parsed_instructions[0:index], self.prompt_engine))
                case 'assistant':
                    self.logger.debug(f"Skip assistant instruction: {instruction}")
                case _:
                    raise ValueError(f"Unknown role: {instruction}")
        self.parsed_instructions = parsed_instructions
        self.logger.debug(self.parsed_instructions)

    def parse_llm(self, string: str, reference: bool = False) -> dict:
        try:
            json_parsed = json.loads(string)
            if json_parsed.get('function_call') is not None:
                # function call generation
                function = json_parsed['function_call']
                parsed = {
                    'type': 'function_call',
                    'name': function['name'],
                    'arguments': json.loads(function['arguments']),
                    'role': 'assistant',
                }
                if reference:
                    self.validation_type = 'function_call'
            else:
                # reservation generation
                status = 'valid'
                content = json_parsed['content']
                if content.endswith('###'):
                    content = content[:-3]
                doubled_content = re.findall(r'(?<=})\n?{.*', content, flags=re.DOTALL)
                if len(doubled_content) > 0:
                    content = doubled_content[-1]
                    status = 'doubled'
                parsed = {
                    'type': 'reservation',
                    'name': 'reservation',
                    'reservation': json.loads(content),
                    'status': status,
                    'role': 'assistant',
                }
                if reference:
                    self.validation_type = 'reservation'
            self.logger.debug(f"Parsing LLM: {parsed}")
            return parsed
        except json.JSONDecodeError as e:
            return { 'status': 'fail', 'error': e }


    def validate(self, **kwargs) -> dict:
        if self.configurations['parse_instruction'] and len(self.parsed_instructions) == 0:
            self.parse_instructions()
        if self.parsed_reference is None:
            self.parsed_reference = self.parse_llm(self.reference, True)
        if self.parsed_generated is None:
            self.parsed_generated = self.parse_llm(self.generated)

        if self.parsed_reference.get('status') == 'fail':
            return { 'passed': False, 'count': {'passed': 0, 'total': 1}, 'status': 'fail', 'source': 'reference', 'error': f"Failed to parse reference (invalid JSON string): {self.parsed_reference['error']}" }
        if self.parsed_generated.get('status') == 'fail':
            return { 'passed': False, 'count': {'passed': 0, 'total': 1}, 'status': 'fail', 'source': 'generated', 'error': f"Failed to parse generated (invalid JSON string): {self.parsed_generated['error']}" }

        if self.validation_type == 'function_call':
            validation_reports = function_call.validate(
                self.parsed_reference['name'],
                self.parsed_generated,
                self.parsed_reference,
                self.parsed_instructions,
                self.configurations,
                self.prompt_engine,
                self.configurations['api'],
                self.logger,
            )
        else:
            validation_reports = reservation.validate(
                self.parsed_generated,
                self.parsed_reference,
                self.parsed_instructions,
                self.configurations,
                self.prompt_engine,
                self.configurations['api'],
                self.logger,
            )
        validation_reports['reference'] = self.parsed_reference
        validation_reports['generated'] = self.parsed_generated
        validation_reports['given_information'] = self.parsed_instructions
        return validation_reports


def __csv_to_list(file_path: str, encoding = 'utf-8', logger = logging.getLogger()) -> list:
    import csv

    llm_list = []
    with open(file_path, mode='r', encoding=encoding) as csv_file:
        csv_reader = csv.DictReader(csv_file)
        line_count = 0
        for row in csv_reader:
            if line_count == 0:
                logger.info(f'Column names are {", ".join(row)}')
                line_count += 1
            llm_list.append(row)
            line_count += 1
        logger.info(f'Processed {line_count} lines.')

    return llm_list

def validate(
    llm_list: Optional[list] = None,
    csv_path: Optional[str] = None,
    columns = {
        'preknowledge': 'context',
        'reference': 'answer',
        'generated': 'generated',
    },
    indexes: Optional[Union[list, range, int]] = None,
    print_result: bool = False,
    logger = logging.getLogger(), **kwargs
) -> dict:
    if isinstance(llm_list, str) and csv_path is None:
        csv_path = llm_list
        llm_list = None
    if llm_list is None:
        if csv_path is None:
            raise ValueError("Either llm_list or csv_path should be provided.")
        llm_list = __csv_to_list(csv_path, kwargs.get('encoding', 'utf-8'), logger)

    if isinstance(indexes, int):
        indexes = [indexes]
    elif isinstance(indexes, range):
        indexes = list(indexes)
    elif indexes is None:
        indexes = range(len(llm_list))
    llms = [llm_list[index] for index in indexes]

    validations = []
    for index, llm in tqdm(zip(indexes, llms), total = len(indexes)):
        validation = EquvalentLLM(
            llm.get(columns['preknowledge']),
            llm.get(columns['reference']),
            llm.get(columns['generated']),
            logger,
            **kwargs
        )
        validated = validation.validate()
        validated['index'] = index
        validations.append(validated)
        if print_result:
            print(validated)

    return {
        'input_data': llm_list,
        'validations': validations,
    }



import json
from logging import Logger

from langchain.prompts import ChatPromptTemplate, SystemMessagePromptTemplate, HumanMessagePromptTemplate
from langchain.output_parsers import PydanticOutputParser
from langchain_openai import ChatOpenAI

from pydantic.v1 import BaseModel, Field

__system_template = '''
- format instructions: {format_instructions}
- When you reserve a movie at movie theater, you should think following considerations.
- DATETIME: time, data, and datetime
- LOCATION': district, place name, landmark, and company name near by theater, and so on. MUST NOT include theater name.
- COORDINATES: coordinates by longitude and latitude
- THEATER': theater name
- ACTOR: well-known actor or actress names
- MOVIE_GENRE: movie genre, which also includes special circumstance like "date" and "relaxation"
- MOVIE_TITLE: specific movie title. Generic name like "영화" MUST NOT belong in this element.
- DECISION: acception or rejection in English or Korean
- RESERVATION: reservation, booking, or cancelation
'''

__human_template = '''
- What kinds of elements are contained below message? Choose the element type at first and describe the evidence of decision.
- MUST exclude elements out of instruction
- MESSAGE: {message}
'''

class UserEntity(BaseModel):
    entity: str = Field(desription="Entity type in user input message")
    keyword: str = Field(description="Core keyword of entity")
    evidence: str = Field(description="Evidence of decision")

class ListOfUserEntity(BaseModel):
    entities: list[UserEntity] = Field(description="List of entities")

def get_entities(message: str, instructions: list, prompt_engine: ChatOpenAI, logger: Logger) -> dict:
    system_template = SystemMessagePromptTemplate.from_template(__system_template)
    human_template = HumanMessagePromptTemplate.from_template(__human_template)
    chat_prompt_template = ChatPromptTemplate.from_messages(
        [system_template, human_template]
    )

    parser = PydanticOutputParser(pydantic_object=ListOfUserEntity)
    final_prompt = chat_prompt_template.format_prompt(
        message=message,
        format_instructions=parser.get_format_instructions()
    ).to_messages()
    logger.debug(f"Prompt for user input message: {final_prompt}")

    result = prompt_engine(final_prompt)
    parsed = json.loads(result.content.split("```json")[1].split("```")[0])
    parsed['role'] = 'user'
    parsed['name'] = 'user'
    parsed['request'] = message
    return parsed



from logging import Logger
from typing import Any

from equivalent_llm import testers
from equivalent_llm.function_call import v1
from equivalent_llm.function_call import v2

def check_function_call(generated: str, reference: str) -> dict:
    if generated != reference:
        return {
            'passed': False,
            'reason': 'Wrong function call',
            'level': 'fatal',
            'evidance': {
                'type': 'function_call',
                'reference': reference,
                'generated': generated,
            }
        }
    else:
        return {'passed': True}

def check_required_parameters(arguments: dict, required: list) -> dict:
    # check if all required parameters are present
    argument_names = arguments.keys()
    for p in required:
        if p not in argument_names:
            return {
                'passed': False,
                'level': 'critical',
                'evidance': {
                    'not_in_required': p,
                },
            }
    return {'passed': True}

def check_paired_arguments(generated_keys: dict, reference_keys: dict) -> dict:
    if generated_keys != reference_keys:
        not_in_reference = list(filter(lambda v: v not in generated_keys, reference_keys))
        not_in_generated = list(filter(lambda v: v not in reference_keys, generated_keys))
        return {
            'passed': False,
            'level': 'severe',
            'evidance': {
                'not_in_reference': not_in_reference,
                'not_in_generated': not_in_generated,
            }
        }
    return {'passed': True}

def equivalence_test(
    generated: Any,
    reference: Any,
    instructions: list,
    definition: dict,
    prompt_engine: Any,
    logger: Logger,
) -> dict:
    result = {}
    # value equality
    if generated != reference:
        get_tester = getattr(testers, definition['tester'])
        tester = get_tester(definition, prompt_engine)
        return tester(generated, reference, instructions, logger)
    return {'passed': True, 'evidence': "exactly mactched"}

def consistency_test(
    generated: Any,
    instructions: list,
    definition: dict,
    prompt_engine: Any,
    logger: Logger,
) -> dict:
    get_tester = getattr(testers, definition['tester'])
    tester = get_tester(definition, prompt_engine)
    return tester(generated, None, instructions, logger)

def grammar_test(
    generated: Any,
    definition: dict,
    prompt_engine: Any,
    logger: Logger,
) -> dict:
    get_tester = getattr(testers, definition['tester'])
    tester = get_tester(definition, prompt_engine)
    return tester(generated, None, None, logger)

def elegance_test(
    generated: Any,
    reference: Any,
    instructions: list,
    definition: dict,
    prompt_engine: Any,
    logger: Logger,
) -> dict:
    get_tester = getattr(testers, definition['tester'])
    tester = get_tester(definition, prompt_engine)
    return tester(generated, reference, instructions, logger)

### function call

def validate(
    function_name: str,
    generated: dict,
    reference: dict,
    instructions: list,
    configurations: dict,
    prompt_engine: Any,
    api_version: str,
    logger: Logger,
) -> dict:
    try:
        match(api_version):
            case 'v1':
                DEFINITION = getattr(v1, function_name.upper())
            case _:
                DEFINITION = getattr(v2, function_name.upper())
    except AttributeError:
        logger.error(f'Function {function_name} not found in equivalent_llm')
        return {'passed': False, 'count': {'total': {'passed': 0, 'total': 1}}, 'tests': {}, 'reason': 'Function not found', 'level': 'fatal'}

    reports = {
        'target': function_name,
        'tests': {
            'equivalence': [],
            'consistency': [],
            'grammar': [],
            'elegance': [],
        },
    }

    if generated['type'] is not reference['type']:
        return {'passed': False, 'count': {'total': {'passed': 0, 'total': 1}}, 'tests': {}, 'reason': 'function call is not generated', 'level': 'fatal'}

    # counters

    equivalence_counter = testers.TestCounter()
    consistency_counter = testers.TestCounter()
    grammar_counter = testers.TestCounter()
    elegance_counter = testers.TestCounter()
    etc_counter = testers.TestCounter()

    ### mandatory tests

    # function name
    function_call_result = check_function_call(generated['name'], reference['name'])
    reports['tests']['function_name'] = function_call_result
    etc_counter.count(function_call_result)

    # required parameters
    required_result = check_required_parameters(generated['arguments'], DEFINITION['required'])
    reports['tests']['required'] = required_result
    etc_counter.count(required_result)

    # argument paried to reference
    paired_arguments_result = check_paired_arguments(generated['arguments'].keys(), reference['arguments'].keys())
    reports['tests']['paired_arguments'] = paired_arguments_result
    etc_counter.count(paired_arguments_result)

    # value type
    for name in generated['arguments']:
        definition = DEFINITION['parameters'].get(name)

        # equivalence tests
        equivalence_result = equivalence_test(
            generated['arguments'].get(name),
            reference['arguments'].get(name),
            instructions,
            definition['equivalence'],
            prompt_engine,
            logger,
        )
        reports['tests']['equivalence'].append({'argument': name, **equivalence_result})
        equivalence_counter.count(equivalence_result)

        ### supplementary tests

        # consistency tests
        if definition.get('consistency'):
            consistency_result = consistency_test(
                generated['arguments'].get(name),
                instructions,
                definition['consistency'],
                prompt_engine,
                logger,
            )
            reports['tests']['consistency'].append({'argument': name, **consistency_result})
            consistency_counter.count(consistency_result)

        # grammar tests
        if definition.get('grammar'):
            grammar_result = grammar_test(
                generated['arguments'].get(name),
                definition['grammar'],
                prompt_engine,
                logger,
            )
            reports['tests']['grammar'].append({'argument': name, **grammar_result})
            grammar_counter.count(grammar_result)

        # grammar tests
        if definition.get('elegance'):
            elegance_result = elegance_test(
                generated['arguments'].get(name),
                reference['arguments'].get(name),
                instructions,
                definition['elegance'],
                prompt_engine,
                logger,
            )
            reports['tests']['elegance'].append({'argument': name, **elegance_result})
            elegance_counter.count(elegance_result)

    total_counter = equivalence_counter + consistency_counter + grammar_counter + elegance_counter

    reports['passed'] = total_counter.fully_passed()
    reports['count'] = {
        'total': total_counter.settlement(),
        'equivalence': equivalence_counter.settlement(),
        'consistency': consistency_counter.settlement(),
        'grammar': grammar_counter.settlement(),
        'elegance': elegance_counter.settlement(),
        'etc': etc_counter.settlement(),
    }
    return reports



import json
import math
from logging import Logger
from typing import Any, Callable, Optional

from langchain.output_parsers import PydanticOutputParser
from langchain.prompts import (
    ChatPromptTemplate,
    SystemMessagePromptTemplate,
    AIMessagePromptTemplate,
    HumanMessagePromptTemplate,
)
from pydantic.v1 import BaseModel, Field

class TestCounter:

    passed: int = 0
    total: int = 0

    def __iadd__(self, counter):
        self.passed += counter.passed
        self.total += counter.total
        return self

    def __add__(self, counter):
        new_counter = TestCounter()
        new_counter.passed = self.passed + counter.passed
        new_counter.total = self.total + counter.total
        return new_counter

    def count(self, validation: dict):
        self.total += 1
        if validation['passed']:
            self.passed += 1

    def fully_passed(self):
        return self.passed == self.total

    def settlement(self):
        return {'passed': self.passed, 'total': self.total}


FormatTemplate = """- format instructions: {format_instructions}"""
PreknowledgeTemplate = """- Related categories: {category}
- given information: {information}"""
EquivalenceTemplate = """- reference sentence: {reference}
- target sentence: {generated}
Question: Is target sentence equivalent to reference sentence?
"""
class EquivalenceResult(BaseModel):
    passed: bool = Field(description="Whether equivalent or not")
    score: int = Field(description="Equivalence score from 0 to 100")
    evidence: str = Field(description="The evidence of passed value")
equivalence_parser = PydanticOutputParser(pydantic_object=EquivalenceResult)

def get_equivalence_tester(definition: dict, prompt_engine: Any) -> Callable:
    # currently fixed to use ChatOpenAI
    def equivalence_tester(value: Any, reference_value: Any, instructions: list, logger: Logger) -> dict:
        try:
            system_template = SystemMessagePromptTemplate.from_template(FormatTemplate)
            ai_template = AIMessagePromptTemplate.from_template(PreknowledgeTemplate)
            human_template = HumanMessagePromptTemplate.from_template(EquivalenceTemplate)
            prompt_template = ChatPromptTemplate.from_messages(
                [system_template, ai_template, human_template]
            )
            final_prompt = prompt_template.format_prompt(
                reference=reference_value,
                generated=value,
                category=definition['category'],
                information=instructions if instructions is not None else "No information",
                format_instructions=equivalence_parser.get_format_instructions(),
            ).to_messages()
            logger.debug(f"Equivalence Prompt: {final_prompt}")
            result = prompt_engine(final_prompt)

            # markdown json format
            if result.content.startswith("```json"):
                return json.loads(result.content.split("```json")[1].split("```")[0])
            else:
                return json.loads(result.content)
        except Exception as e:
            logger.debug(f"Equvalence Error: {e}")
            return {
                'passed': False,
                'level': 'fatal',
                'evidence': {
                    'type': 'validation process',
                    'message': str(e),
                }
            }
    return equivalence_tester

# consistency testers

LatestRequestedTemplate = """- Given information: {given_information}"""
NumberConsistencyTemplate = """- Answer a question by true or false of passed with concrete evidence.
- Question: Is below a value is consistently matched with given information? {range_sentence}
{comment}
- target value: {generated}
"""

def get_number_consistency_tester(definition: dict, prompt_engine: Any) -> Callable:
    def number_tester(value: Any, reference_value: Any, instructions: list, logger: Logger) -> dict:
        range_sentence = ''
        if definition.get('range') is not None:
            if definition['range'][0] == -math.inf:
                range_sentence = f"Also, is the value less or equal than {definition['range'][1]}?"
            elif definition['range'][1] == math.inf:
                range_sentence = f"Also, is the value greater or equal than {definition['range'][0]}?"
            else:
                range_sentence = f"Also, is the value between {definition['range'][0]} and {definition['range'][1]}?"
        logger.debug(f"Range sentence: {range_sentence}")
        try:
            system_template = SystemMessagePromptTemplate.from_template(FormatTemplate)
            ai_template = AIMessagePromptTemplate.from_template(LatestRequestedTemplate)
            human_template = HumanMessagePromptTemplate.from_template(NumberConsistencyTemplate)
            prompt_template = ChatPromptTemplate.from_messages(
                [system_template, ai_template, human_template]
            )
            final_prompt = prompt_template.format_prompt(
                generated=value,
                given_information=instructions if instructions is not None else "No information",
                range_sentence=range_sentence,
                comment=definition.get('comment', ''),
                format_instructions=equivalence_parser.get_format_instructions(),
            ).to_messages()
            logger.debug(f"Number Consistency Prompt: {final_prompt}")
            result = prompt_engine(final_prompt)

            # markdown json format
            if result.content.startswith("```json"):
                return json.loads(result.content.split("```json")[1].split("```")[0])
            else:
                return json.loads(result.content)
        except Exception as e:
            logger.debug(f"Number Consistency Error: {e}")
            return {
                'passed': False,
                'level': 'fatal',
                'evidence': {
                    'type': 'validation process',
                    'message': str(e),
                }
            }
    return number_tester

CategoryConsistencyTemplate = """- Answer a question by true or false of passed with concrete evidence.
- Question: Is below phrase is matched with given information of following categories: {category_sentence}?
  {comment}
- target sentence: {generated}
"""

def get_category_consistency_tester(definition: dict, prompt_engine: Any) -> Callable:
    def category_tester(value: Any, reference_value: Any, instructions: list, logger: Logger) -> dict:
        try:
            system_template = SystemMessagePromptTemplate.from_template(FormatTemplate)
            ai_template = AIMessagePromptTemplate.from_template(LatestRequestedTemplate)
            human_template = HumanMessagePromptTemplate.from_template(CategoryConsistencyTemplate)
            prompt_template = ChatPromptTemplate.from_messages(
                [system_template, ai_template, human_template]
            )
            final_prompt = prompt_template.format_prompt(
                generated=value,
                given_information=instructions if instructions is not None else "No information",
                category_sentence=definition['category'],
                comment=definition.get('comment', ''),
                format_instructions=equivalence_parser.get_format_instructions(),
            ).to_messages()
            logger.debug(f"Category Consistency Prompt: {final_prompt}")
            result = prompt_engine(final_prompt)
            # markdown json format
            if result.content.startswith("```json"):
                return json.loads(result.content.split("```json")[1].split("```")[0])
            else:
                return json.loads(result.content)
        except Exception as e:
            logger.debug(f"Category Consistency Error: {e}")
            return {
                'passed': value == reference_value,
                'evidence': {
                    'type': 'validation process',
                    'message': str(e),
                }
            }
    return category_tester

EnumConsistencyTemplate = """- Answer a question by true or false of passed with concrete evidence.
- Question: Is below phrase is matched with given enum items: {enum_items}?
  {comment}
- target sentence: {generated}
"""

def get_enum_consistency_tester(definition: dict, prompt_engine: Any) -> Callable:
    def enum_tester(value: Any, reference_value: Any, instructions: list, logger: Logger) -> dict:
        try:
            system_template = SystemMessagePromptTemplate.from_template(FormatTemplate)
            ai_template = AIMessagePromptTemplate.from_template(LatestRequestedTemplate)
            human_template = HumanMessagePromptTemplate.from_template(EnumConsistencyTemplate)
            prompt_template = ChatPromptTemplate.from_messages(
                [system_template, ai_template, human_template]
            )
            final_prompt = prompt_template.format_prompt(
                generated=value,
                given_information=instructions if instructions is not None else "No information",
                enum_items=definition['itmes'],
                comment=definition.get('comment', ''),
                format_instructions=equivalence_parser.get_format_instructions(),
            ).to_messages()
            logger.debug(f"Enum Consistency Prompt: {final_prompt}")
            result = prompt_engine(final_prompt)
            # markdown json format
            if result.content.startswith("```json"):
                return json.loads(result.content.split("```json")[1].split("```")[0])
            else:
                return json.loads(result.content)
        except Exception as e:
            logger.debug(f"Enum Consistency Error: {e}")
            return {
                'passed': value == reference_value,
                'evidence': {
                    'type': 'validation process',
                    'message': str(e),
                }
            }
    return enum_tester

GrammarTemplate = """- You are a experienced grammarian.
- Answer a question by true or false of passed with concrete evidence.
- Question: Is below sentence is composed with STRICTLY correct grammar for Korean, English, or mixture of Korean and English, is it acceptable to read naturally?
  If sentence is just a word, it is considered as correct.
  {comment}
- target sentence: {generated}
"""

def get_grammar_tester(definition: dict, prompt_engine: Any) -> Callable:
    def grammar_tester(value: Any, reference_value: Any, instructions: Optional[list], logger: Logger) -> dict:
        try:
            system_template = SystemMessagePromptTemplate.from_template(FormatTemplate)
            human_template = HumanMessagePromptTemplate.from_template(GrammarTemplate)
            prompt_template = ChatPromptTemplate.from_messages(
                [system_template, human_template]
            )
            final_prompt = prompt_template.format_prompt(
                generated=value,
                comment=definition.get('comment', ''),
                format_instructions=equivalence_parser.get_format_instructions(),
            ).to_messages()
            logger.debug(f"Grammar Prompt: {final_prompt}")
            result = prompt_engine(final_prompt)

            # markdown json format
            if result.content.startswith("```json"):
                return json.loads(result.content.split("```json")[1].split("```")[0])
            else:
                return json.loads(result.content)
        except Exception as e:
            logger.debug(f"Grammar Error: {e}")
            return {
                'passed': value == reference_value,
                'evidence': {
                    'type': 'validation process',
                    'message': str(e),
                }
            }
    return grammar_tester

class EleganceResult(BaseModel):
    passed: bool = Field(description="Whether equivalent or not")
    score: int = Field(description="Equivalence score from 0 to 100")
    evidence: str = Field(description="The evidence of passed value")
    alternative: str = Field(description="Alternative sentence")
elegance_parser = PydanticOutputParser(pydantic_object=EleganceResult)

PreknowledgeTemplate = """- given information: {information}"""
EleganceTemplate = """- reference sentence: {reference}
- Question: Is below target sentence centainly superior elegant than, reference sentence?
  Please provide a elegant and human readable alternative as well by orignial language of target sentence with same tone.
  {comment}
- target sentence: {generated}
"""

"Is the sentence below equivalent to, or even more elegant than, the standard sentence? Please provide a fluid and refined alternative as well."
def get_elegance_tester(definition: dict, prompt_engine: Any) -> Callable:
    # currently fixed to use ChatOpenAI
    def elegance_tester(value: Any, reference_value: Any, instructions: list, logger: Logger) -> dict:
        try:
            system_template = SystemMessagePromptTemplate.from_template(FormatTemplate)
            ai_template = AIMessagePromptTemplate.from_template(PreknowledgeTemplate)
            human_template = HumanMessagePromptTemplate.from_template(EleganceTemplate)
            prompt_template = ChatPromptTemplate.from_messages(
                [system_template, ai_template, human_template]
            )
            final_prompt = prompt_template.format_prompt(
                reference=reference_value,
                generated=value,
                comment=definition.get('comment', ''),
                information=instructions if instructions is not None else "No information",
                format_instructions=elegance_parser.get_format_instructions(),
            ).to_messages()
            logger.debug(f"Elegance Prompt: {final_prompt}")
            result = prompt_engine(final_prompt)

            # markdown json format
            if result.content.startswith("```json"):
                return json.loads(result.content.split("```json")[1].split("```")[0])
            else:
                return json.loads(result.content)
        except Exception as e:
            logger.debug(f"Elegance Error: {e}")
            return {
                'passed': False,
                'level': 'fatal',
                'evidence': {
                    'type': 'validation process',
                    'message': str(e),
                }
            }
    return elegance_tester



from logging import Logger
from typing import Any, Optional

from equivalent_llm import testers

RESERVATION_BOARD = {
    'answer': {
        'name': 'answer',
        'equivalence': {
            'tester': 'get_equivalence_tester',
            'category': 'answer for a question or a request of user, suggestion of related to movie reservation, or confirmation of reservation',
        },
        'consistency': {
            'tester': 'get_category_consistency_tester',
            'instruction': 'all',
            'category': 'answer for a question or a request of user, suggestion of related to movie reservation, or confirmation of reservation',
        },
        'grammar': {
            'tester': 'get_grammar_tester',
        },
        'elegance': {
            'tester': 'get_elegance_tester',
        },
    },
    'selected_title': {
        'name': 'selected_title',
        'equivalence': {
            'tester': 'get_equivalence_tester',
            'category': 'movie title',
        },
        'consistency': {
            'tester': 'get_category_consistency_tester',
            'instruction': 'all',
            'category': 'movie title',
        },
    },
    'selected_theater': {
        'name': 'selected_theater',
        'equivalence': {
            'tester': 'get_equivalence_tester',
            'category': 'theater name',
        },
        'consistency': {
            'tester': 'get_category_consistency_tester',
            'instruction': 'all',
            'category': 'theater name',
        },
    },
    'selected_schedule': {
        'name': 'selected_schedule',
        'equivalence': {
            'tester': 'get_equivalence_tester',
            'category': 'schedule',
        },
        'consistency': {
            'tester': 'get_category_consistency_tester',
            'instruction': 'all',
            'category': 'schedule',
        },
    },
    'selected_time': {
        'name': 'selected_time',
        'equivalence': {
            'tester': 'get_equivalence_tester',
            'category': 'time',
        },
        'consistency': {
            'tester': 'get_category_consistency_tester',
            'instruction': 'all',
            'category': 'time',
        },
    },
    'reference_number': {
        'name': 'reference_number',
        'equivalence': {
            'tester': 'get_equivalence_tester',
            'category': 'ref:DDD, ref:DDDD',
        },
    },
    'template': {
        'name': 'template',
        'equivalence': {
            'tester': 'get_equivalence_tester',
            'category': 'template of theater or movies, none',
        },
    },
}

def equivalence_test(
    generated: Any,
    reference: Any,
    instructions: Optional[list],
    definition: dict,
    prompt_engine: Any,
    logger: Logger,
) -> dict:
    if generated != reference:
        get_tester = getattr(testers, definition['tester'])
        tester = get_tester(definition, prompt_engine)
        return tester(generated, reference, instructions, logger)
    return {'passed': True, 'evidence': "exactly matched"}

def consistency_test(
    generated: Any,
    instructions: Optional[list],
    definition: dict,
    prompt_engine: Any,
    logger: Logger,
) -> dict:
    get_tester = getattr(testers, definition['tester'])
    tester = get_tester(definition, prompt_engine)
    return tester(generated, None, instructions, logger)

def grammar_test(
    generated: Any,
    definition: dict,
    prompt_engine: Any,
    logger: Logger,
) -> dict:
    get_tester = getattr(testers, definition['tester'])
    tester = get_tester(definition, prompt_engine)
    return tester(generated, None, None, logger)

def elegance_test(
    generated: Any,
    reference: Any,
    instructions: Optional[list],
    definition: dict,
    prompt_engine: Any,
    logger: Logger,
) -> dict:
    get_tester = getattr(testers, definition['tester'])
    tester = get_tester(definition, prompt_engine)
    return tester(generated, reference, instructions, logger)

def validate(
    generated: dict,
    reference: dict,
    instructions: Optional[list],
    configurations: dict,
    prompt_engine: Any,
    api_version: str,
    logger: Logger,
) -> dict:
    reports = {
        'target': 'reservation_board',
        'tests': {
            'equivalence': [],
            'consistency': [],
            'grammar': [],
            'elegance': [],
        },
    }

    elements = [
        'answer',
        'selected_title',
        'selected_theater',
        'selected_schedule',
        'selected_time',
        'reference_number',
        'template',
    ]

    if generated['type'] is not reference['type']:
        return {'passed': False, 'count': {'total': {'passed': 0, 'total': 1}}, 'tests': {}, 'reason': 'reservation is not generated ', 'level': 'fatal'}

    equivalence_counter = testers.TestCounter()
    consistency_counter = testers.TestCounter()
    grammar_counter = testers.TestCounter()
    elegance_counter = testers.TestCounter()

    for element in elements:
        if (
            RESERVATION_BOARD.get(element) and
            (
                (
                    generated['reservation'].get(element) is not None and
                    generated['reservation'][element] != ''
                ) or (
                    reference['reservation'].get(element) is not None and
                    reference['reservation'][element] != ''
                )
            )
        ) :
            # equivalence test
            equivalence_result = equivalence_test(
                generated['reservation'].get(element),
                reference['reservation'].get(element),
                instructions,
                RESERVATION_BOARD[element]['equivalence'],
                prompt_engine,
                logger,
            )
            reports['tests']['equivalence'].append({'element': element, **equivalence_result})
            equivalence_counter.count(equivalence_result)

            # consistency test
            if RESERVATION_BOARD[element].get('consistency'):
                consistency_result = consistency_test(
                    generated['reservation'].get(element),
                    instructions,
                    RESERVATION_BOARD[element]['consistency'],
                    prompt_engine,
                    logger,
                )
                reports['tests']['consistency'].append({'element': element, **consistency_result})
                consistency_counter.count(consistency_result)

            # grammar test
            if RESERVATION_BOARD[element].get('grammar'):
                grammar_result = grammar_test(
                    generated['reservation'].get(element),
                    RESERVATION_BOARD[element]['grammar'],
                    prompt_engine,
                    logger,
                )
                reports['tests']['grammar'].append({'element': element, **grammar_result})
                grammar_counter.count(grammar_result)

            # elegance test
            if RESERVATION_BOARD[element].get('elegance'):
                elegance_result = elegance_test(
                    generated['reservation'].get(element),
                    reference['reservation'].get(element),
                    instructions,
                    RESERVATION_BOARD[element]['elegance'],
                    prompt_engine,
                    logger,
                )
                reports['tests']['elegance'].append({'element': element, **elegance_result})
                elegance_counter.count(elegance_result)

    total_counter = equivalence_counter + consistency_counter + grammar_counter + elegance_counter

    reports['passed'] = total_counter.fully_passed()
    reports['count'] = {
        'total': total_counter.settlement(),
        'equivalence': equivalence_counter.settlement(),
        'consistency': consistency_counter.settlement(),
        'grammar': grammar_counter.settlement(),
        'elegance': elegance_counter.settlement(),
    }
    return reports
